from qt import QObject, PYSIGNAL

__all__ = ['getProxy']

class _TableRowProxy(object):
    """
    self.num is a row index in the table
    """
    def __setitem__(self,i,v):
        if type(i) == str:
            i = self.getColumnIndex(i)
        super(self.__class__,self).__setitem__(i,v)
        self.tab.emitter.emit(PYSIGNAL("cellChanged"),(self.num,i,self.data[i]))

    def select(self):
        self.tab.select(self.num)
        
class _TableModelProxy(object):
    def __init__(self, *args):
        self.super = super(self.__class__,self)
        self.super.__init__(*args)
        self.emitter = QObject()
        self.selection = None
        
    def setHeader(self, col, header):
        self.super.setHeader(col, header)
        self.emitter.emit(PYSIGNAL("setHeader"),(col,header))
        
    def insertRow(self,i=None,row=None):
        if i is None: i=len(self.table)
        if self.super.insertRow(i,row):
            for k in range(i,len(self.table)):
                self.table[k].num = k
            self.emitter.emit(PYSIGNAL("insertRow"),(i,self.table[i]))
            return True
        else:
            return False
        
    def insertColumn(self,i=None,col=None):
        if self.super.insertColumn(i,col):
            self.emitter.emit(PYSIGNAL("insertColumn"),(i,col))
            if type(i)==int and i <= self.selection:
                self.selection += 1
            return True
        else:
            return False
    
    def takeRow(self,i):
        r = self.super.takeRow(i)
        for k in range(i,len(self.table)):
            self.table[k].num = k
        if i == self.selection:
            self.selection = None
        self.emitter.emit(PYSIGNAL("takeRow"),(i,))
        return r
    
    def takeColumn(self,i):
        c = self.super.takeColumn(i)
        self.emitter.emit(PYSIGNAL("takeColumn"),(i,))
        if i == self.selection:
            self.selection = None
        elif i < self.selection:
            self.selection -= 1
        return c
    
    def sort(self, *args):
        self.super.sort(*args)
        for i,row in enumerate(self.table):
            row.num = i
        self.emitter.emit(PYSIGNAL("sort"),())

    def select(self, sel):
        self.selection = sel
        self.emitter.emit(PYSIGNAL("select"),(sel,))

    def getSelection(self):
        return self.selection
    
def getProxy(cls):
    name = "Proxy_TableRow"
    bases = (cls.TableRow,) + _TableRowProxy.__bases__
    dic = dict(_TableRowProxy.__dict__)
    rowModelClass = type(name, bases, dic)
    
    name = "Proxy_" + cls.__name__
    bases = (cls,) + _TableModelProxy.__bases__
    dic = dict(_TableModelProxy.__dict__)
    dic['TableRow'] = rowModelClass
    
    return type(name, bases, dic)

